/****************************************************************************
**
** Copyright (C) 2016 The Qt Company Ltd.
** Contact: https://www.qt.io/licensing/
**
** This file is part of Qbs.
**
** $QT_BEGIN_LICENSE:LGPL$
** Commercial License Usage
** Licensees holding valid commercial Qt licenses may use this file in
** accordance with the commercial license agreement provided with the
** Software or, alternatively, in accordance with the terms contained in
** a written agreement between you and The Qt Company. For licensing terms
** and conditions see https://www.qt.io/terms-conditions. For further
** information use the contact form at https://www.qt.io/contact-us.
**
** GNU Lesser General Public License Usage
** Alternatively, this file may be used under the terms of the GNU Lesser
** General Public License version 3 as published by the Free Software
** Foundation and appearing in the file LICENSE.LGPL3 included in the
** packaging of this file. Please review the following information to
** ensure the GNU Lesser General Public License version 3 requirements
** will be met: https://www.gnu.org/licenses/lgpl-3.0.html.
**
** GNU General Public License Usage
** Alternatively, this file may be used under the terms of the GNU
** General Public License version 2.0 or (at your option) the GNU General
** Public license version 3 or any later version approved by the KDE Free
** Qt Foundation. The licenses are as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL2 and LICENSE.GPL3
** included in the packaging of this file. Please review the following
** information to ensure the GNU General Public License requirements will
** be met: https://www.gnu.org/licenses/gpl-2.0.html and
** https://www.gnu.org/licenses/gpl-3.0.html.
**
** $QT_END_LICENSE$
**
****************************************************************************/
#ifndef COMMANDLINEPARSER_H
#define COMMANDLINEPARSER_H

#include "configcommand.h"

#include <tools/settings.h>

#include <QtCore/qstringlist.h>

class ConfigCommandLineParser
{
public:
    void parse(const QStringList &commandLine);

    ConfigCommand command() const { return m_command; }

    qbs::Settings::Scopes scope() const { return m_scope; }
    QString settingsDir() const { return m_settingsDir; }
    bool helpRequested() const { return m_helpRequested; }
    void printUsage() const;

    class Error
    {
    public:
        Error(QString message) : m_message(std::move(message)) { }
        QString message() const { return m_message; }
    private:
        QString m_message;
    };

private:
    void assignOptionArgument(const QString &option, QString &argument);
    void setCommand(ConfigCommand::Command command);
    void setScope(qbs::Settings::Scope scope);

    ConfigCommand m_command;
    qbs::Settings::Scopes m_scope = qbs::Settings::allScopes();
    bool m_helpRequested = false;
    QString m_settingsDir;
    QStringList m_commandLine;
};

#endif // COMMANDLINEPARSER_H
