//![0]
#include <lib.h>
#include <version.h>

#include <stdio.h>

int main()
{
    printf("Hello, world\n");
    printf("%s\n", get_string());
    printf("ProductVersion = %s\n", kProductVersion);
    return 0;
}
//![0]
