/****************************************************************************
**
** Copyright (C) 2016 The Qt Company Ltd.
** Copyright (C) 2019 Jochen Ulrich <jochenulrich@t-online.de>
** Contact: https://www.qt.io/licensing/
**
** This file is part of Qbs.
**
** $QT_BEGIN_LICENSE:LGPL$
** Commercial License Usage
** Licensees holding valid commercial Qt licenses may use this file in
** accordance with the commercial license agreement provided with the
** Software or, alternatively, in accordance with the terms contained in
** a written agreement between you and The Qt Company. For licensing terms
** and conditions see https://www.qt.io/terms-conditions. For further
** information use the contact form at https://www.qt.io/contact-us.
**
** GNU Lesser General Public License Usage
** Alternatively, this file may be used under the terms of the GNU Lesser
** General Public License version 3 as published by the Free Software
** Foundation and appearing in the file LICENSE.LGPL3 included in the
** packaging of this file. Please review the following information to
** ensure the GNU Lesser General Public License version 3 requirements
** will be met: https://www.gnu.org/licenses/lgpl-3.0.html.
**
** GNU General Public License Usage
** Alternatively, this file may be used under the terms of the GNU
** General Public License version 2.0 or (at your option) the GNU General
** Public license version 3 or any later version approved by the KDE Free
** Qt Foundation. The licenses are as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL2 and LICENSE.GPL3
** included in the packaging of this file. Please review the following
** information to ensure the GNU General Public License requirements will
** be met: https://www.gnu.org/licenses/gpl-2.0.html and
** https://www.gnu.org/licenses/gpl-3.0.html.
**
** $QT_END_LICENSE$
**
****************************************************************************/

#include "abstractcommandexecutor.h"

#include "rulecommands.h"
#include "transformer.h"

#include <language/language.h>
#include <logging/translator.h>
#include <tools/error.h>

namespace qbs {
namespace Internal {

AbstractCommandExecutor::AbstractCommandExecutor(Logger logger, QObject *parent)
    : QObject(parent)
    , m_echoMode(defaultCommandEchoMode())
    , m_command(nullptr)
    , m_transformer(nullptr)
    , m_mainThreadScriptEngine(nullptr)
    , m_dryRun(false)
    , m_logger(std::move(logger))
{
    m_watchdog.setSingleShot(true);
    connect(&m_watchdog, &QTimer::timeout, this, [this]() {
        cancel(ErrorInfo{Tr::tr("Command cancelled because it exceeded the timeout: %1")
                             .arg(m_command->descriptionForCancelMessage(
                                 m_transformer->product()->fullDisplayName()))});
    });
    connect(this, &AbstractCommandExecutor::finished, &m_watchdog, &QTimer::stop);
}

void AbstractCommandExecutor::start(Transformer *transformer, AbstractCommand *cmd)
{
    m_transformer = transformer;
    m_command = cmd;
    doSetup();
    doReportCommandDescription(transformer->product()->fullDisplayName());
    if (doStart())
        startTimeout();
}

void AbstractCommandExecutor::doReportCommandDescription(const QString &productName)
{
    if (m_command->isSilent() || m_echoMode == CommandEchoModeSilent)
        return;

    if (m_command->description().isEmpty()) {
        m_logger.printWarning(
                    ErrorInfo(Tr::tr("Command is not marked silent, but has no description."),
                              m_command->codeLocation()));
    } else {
        emit reportCommandDescription(m_command->highlight(),
                                      m_command->fullDescription(productName));
    }
}

void AbstractCommandExecutor::startTimeout()
{
    if (!m_dryRun || m_command->ignoreDryRun()) {
        const auto timeout = m_command->timeout();
        if (timeout > 0)
            m_watchdog.start(timeout * 1000);
    }
}

} // namespace Internal
} // namespace qbs
